package botworld;

/* The gnu.getopt package. */
import gnu.getopt.LongOpt;
import gnu.getopt.Getopt;

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;

import prolog.*;
import prolog.util.*;

/**
 * Botworld Applet.
 * <P>
 * Implementation based on the tcl/tk version by
 * <A HREF="MailTo:claude@cse.unsw.edu.au">Claude Sammut</A> and
 * <A HREF="MailTo:malcolmr@cse.unsw.edu.au">Malcolm Ryan</A>
 * and another applet by Claude.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Botworld.java,v 1.4 2000/07/18 03:02:53 peteg Exp $
 */
public class Botworld extends Applet {
    /* Standalone constants. */
    private final static int WIDTH  = 400;
    private final static int HEIGHT = 350;

    Thread[] threads;

    Vector staticObjects = new Vector();
    Vector movableObjects = new Vector();
    Vector worldObjects = new Vector();

    Vector bars = new Vector();

    private static final String PdbURL = "dbURL";
    private static final String PnumBots = "numBots";
    private static final String PworldURL = "worldURL";

    /* Off-screen buffer stuff. */
    Dimension dim;
    private Graphics offGraphics;
    private Image offImage;

    private boolean applet = false;

    public Botworld() {
	super();
    }

    public Dimension getMinimumSize() {
        return(getPreferredSize());
    }

    public Dimension getPreferredSize() {
        return(new Dimension(WIDTH, HEIGHT));
    }

    public int getHeight() {
	return(dim.height);
    }

    public int getWidth() {
	return(dim.width);
    }

    /**
     * Setup for the application.
     */
    void setup(Database db, int numBots, String world, boolean time) {
	/* Get the world dimensions. */
	dim = getSize();

 	try {
	    /* Load the world file, init static objects. */
	    BufferedReader in = new BufferedReader(new FileReader(world));
	    parseWorld(in);
	    in.close();
	} catch (IOException e) {
	    e.printStackTrace();
	    throw new RuntimeException("Botworld.<init>: error loading: " + world + " (IOException)");
	}

	/* Init the bots. */
	threads = new Thread[numBots];
	for(int i = 0; i < numBots; i++) {
	    Bot b = new Bot(i + 1,
			    (i < numBots - 1) ?  (Database)db.clone() : db,
			    this,
			    time);
	    movableObjects.addElement(b);
	    worldObjects.addElement(b);
	    threads[i] = new Thread(b);
	}

	/* Setup the graphics buffer. */
        offImage = createImage(dim.width, dim.height);
	offGraphics = offImage.getGraphics();
    }

    public String getAppletInfo() {
	return("Botworld Applet - Peter Gammie, 1999.");
    }

    public String[][] getParameterInfo() {
	String info[][] = {
	    {PdbURL,	"URL",		"URL of the serialised database of the bot to run."},
	    {PnumBots,	"int",		"Number of bots to run."},
	    {PworldURL,	"URL",		"URL of the text file describing the world."}
	};

	return(info);
    }
   
    /**
     * Parses the parameters, loads the serialized database.
     */
    public void init() {
	int numBots;
	String dbURL, worldURL;
	String s;
	URL documentBase = getDocumentBase();

	/* Get the world dimensions. */
	dim = getSize();

	/* We're invoked as an applet. */
	applet = true;

	if((dbURL = getParameter(PdbURL)) == null)
	    throw new RuntimeException("Botworld.init(): must specify: " + PdbURL);

	numBots = ((s = getParameter(PnumBots)) != null) ? Integer.parseInt(s) : 1;

	if((worldURL = getParameter(PworldURL)) == null)
	    throw new RuntimeException("Botworld.init(): must specify: " + PworldURL);

 	try {
	    /* Fetch the world file, init static objects. */
	    showStatus("Loading world file.");

	    URL url = new URL(documentBase, worldURL);
	    BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));

	    parseWorld(in);
	    in.close();
	} catch (IOException e) {
	    e.printStackTrace();
	    throw new RuntimeException("Botworld.init(): error loading: " + worldURL + " (IOException)");
	}

	try {
	    /* Fetch the database, init the bots. */
	    showStatus("Loading Prolog database.");

	    URL url = new URL(documentBase, dbURL);
	    ObjectInputStream in = new ObjectInputStream(url.openStream());
	    Database db = (Database)in.readObject();

	    threads = new Thread[numBots];
	    for(int i = 0; i < numBots; i++) {
		Bot b = new Bot(i + 1,
				(i < numBots - 1) ?  (Database)db.clone() : db,
				this,
				false);
		movableObjects.addElement(b);
		worldObjects.addElement(b);
		threads[i] = new Thread(b);
	    }

	    in.close();
	} catch (IOException e) {
	    e.printStackTrace();
	    throw new RuntimeException("Botworld.init(): error loading: " + dbURL + " (IOException)");
	} catch (ClassNotFoundException e) {
	    System.err.print("Botworld(database): class not found: ");
	    e.printStackTrace();
	    throw new RuntimeException("Botworld.init(): ClassNotFoundException.");
	}

	/* Setup the graphics buffer. */
        offImage = createImage(dim.width, dim.height);
	offGraphics = offImage.getGraphics();
    }

    /**
     * Parses a world. A world is defined by the grammar:
     * <BR>
     * <PRE>
     * world  -> object world
     * world  -> object
     *
     * object -> bar <num> <xcoord> <ycoord> <angle>
     * object -> beacon <num> <xcoord> <ycoord>
     * object -> box <num> <lxcoord> <tycoord> <rxcoord> <bycoord>
     * </PRE>
     */
    private void parseWorld(Reader in) throws IOException {
	StreamTokenizer st = new StreamTokenizer(in);
	String id;
	int[] args = new int[4];
	WorldObject wo;

	/* Setup the tokenizer. */
        st.commentChar('#');
	st.eolIsSignificant(false);
	st.lowerCaseMode(true);
	st.parseNumbers();
	st.slashStarComments(true);

// 	System.err.println("parseWorld()");

    scan: while(st.nextToken() != StreamTokenizer.TT_EOF) {
//  	System.err.println(st.sval);

	/* Read the identifier (bar/beacon/box) */
	if(st.ttype != StreamTokenizer.TT_WORD)
	    break scan;
	id = st.sval;

	/* Read in three args. */
	for(int i = 0; i < 3; i++) {
	    if(st.nextToken() != StreamTokenizer.TT_NUMBER)
		break scan;
	    args[i] = (int)st.nval;
	}

	/* Create new world object. */
	if(id.equals("bar")) {
	    if(st.nextToken() != StreamTokenizer.TT_NUMBER)
		break scan;
	    wo = new Bar(args[0], args[1], args[2], (int)st.nval);
	    movableObjects.addElement(wo);
	    bars.addElement(wo);
	} else if(id.equals("beacon")) {
	    staticObjects.addElement(wo = new Beacon(args[0], args[1], args[2]));
	} else if(id.equals("box")) {
	    if(st.nextToken() != StreamTokenizer.TT_NUMBER)
		break scan;
	    int arg4 = (int)st.nval;
	    if(st.nextToken() != StreamTokenizer.TT_NUMBER)
		break scan;
	    staticObjects.addElement(wo = new Box(args[0], args[1], args[2], arg4, (int)st.nval));
	} else {
	    break scan;
	}
	worldObjects.addElement(wo);
    }
    }

    /**
     * Kills the bot threads.
     */
    public void destroy() {
	for(int i = 0; i < threads.length; i++)
	    if(threads[i] != null) {
// 		threads[i].stop();
		threads[i] = null;
	    }
    }

    /**
     * Fires up the bot threads.
     */
    public void start() {
	for(int i = 0; i < threads.length; i++) {
	    if(threads[i].isAlive()) {
		threads[i].resume();
	    } else {
		threads[i].start();
	    }
	}

	if(applet) {
	    showStatus("Botworld running.");
	}
    }

    /**
     * Suspends the bot threads.
     */
    public void stop() {
	/* "suspend()" is dodgy under Netscape (at least), so do nothing. */
// 	for(int i = 0; i < threads.length; i++)
// 	    threads[i].suspend();
// 	showStatus(null);
    }

    /**
     * Checks for collisions between objects.
     */
    public synchronized boolean collided(WorldObject obj) {
	int size = worldObjects.size();

	for(int i = 0; i < size; i++) {
	    WorldObject wo = (WorldObject)worldObjects.elementAt(i);

	    if(wo == obj || wo instanceof Bar)
		continue;
	    else
		if(obj.collision(wo))
		    return(true);
	}
	return(false);
    }

    public void paint(Graphics g) {
	update(g);
    }

    public void update(Graphics g) {
	offGraphics.setColor(Color.gray);
	offGraphics.fillRect(0, 0, dim.width, dim.height);

	for(int i = 0; i < worldObjects.size(); i++)
	    ((WorldObject)worldObjects.elementAt(i)).draw(offGraphics);

	g.drawImage(offImage, 0, 0, this);
    }

    /**
     * Standalone entry point. Creates a new graphics context and prolog
     * thread, then drops into the read-eval loop.
     */
    public static void main(String[] argv) {
	String database = null;
	boolean jit = false;
	int numBots = 1;
	boolean time = false;
	String world = "world0";

	Database db;

	/* Print an obnoxious header. */
	System.err.println("Botworld 0.5\n" +
			   "(C)opyright 2000 Peter Gammie.\n" +
			   "Botworld is free software, covered by the GNU Library General Public License,\n" +
			   "and you are welcome to change it and/or distribute copies of it under certain\n" +
			   "conditions.");

	/* Parse the command tail. */
	LongOpt[] longopts = {
	    new LongOpt("benchmark", LongOpt.NO_ARGUMENT, null, 'b'),
	    new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h'),
	    new LongOpt("jit", LongOpt.NO_ARGUMENT, null, 'j'),
	    new LongOpt("loaddb", LongOpt.REQUIRED_ARGUMENT, null, 'l'),
	    new LongOpt("numBots", LongOpt.REQUIRED_ARGUMENT, null, 'n'),
	    new LongOpt("world", LongOpt.REQUIRED_ARGUMENT, null, 'w'),
	};

	Getopt g = new Getopt("botworld", argv, "bhjl:n:w:", longopts);
	int c;
	String arg;
	g.setOpterr(false);

	while((c = g.getopt()) != -1) {
	    switch(c) {
	    case 'b':
		System.err.println("Benchmark option set.");
		time = true;
		break;
	    case 'h':
		System.err.println("\nUsage: botworld.Botworld [OPTION]... [FILE]...\n\n" +
				   "  -b, --benchmark\tTime each top-level query.\n" +
				   "  -h, --help\t\tDisplay this help and exit.\n" +
				   "  -j, --jit\t\tEnable the just-in-time prolog compiler.\n" +
				   "  -l, --loaddb FILE\tLoad a serialized database.\n" +
				   "  -n, --numBots\t\tThe number of bot to create (default: 1).\n" +
				   "  -w, --world\t\t1The world file to use (default: world).\n");
		    System.exit(1);
	    case 'j':
		System.err.println("JIT option set.");
		jit = true;
		break;
	    case 'l':
		database = g.getOptarg();
		break;
	    case 'n':
		arg = g.getOptarg();
		try {
		    numBots = Integer.parseInt(arg);
		} catch (NumberFormatException e) {
		    System.err.println("botworld: invalid parameter to numBots: " + arg);
		    System.err.println("Try `botworld.Botworld --help' for more information.");
		    System.exit(1);
		}
		break;
	    case 'w':
		world = g.getOptarg();
		break;
	    case '?':
		System.err.println("prolog: invalid option " + (char)g.getOptopt());
		System.err.println("Try `prolog --help' for more information.");
		System.exit(1);
	    default:
		System.err.print("getopt() returned " + c + "\n");
		System.exit(1);
	    }
	}

	/* Create the prolog thread given the database specs. */
	if(database != null) {
	    /* Deserialize the database. */
	    System.err.println("Loading serialized database: " + database);
	    try {
		/* Standard technique - straight out of "Java in a Nutshell" */
		FileInputStream fis = new FileInputStream(database);
		ObjectInputStream in = new ObjectInputStream(fis);
		db = (Database)in.readObject();
	    } catch (Exception e) {
		System.err.print(e);
		e.printStackTrace();
		System.exit(1);
		return; // borken control flow analysis.
	    }
	} else {
	    db = new Database(jit);
	}

	/* Read all filenames specified -- they're "non-option"s. */
// 	try {
// 	    for(int i = g.getOptind(); i < argv.length; i++) {
// 		System.err.print("Loading: " + argv[i]);
// 		pt.consult(argv[i]);
// 		System.err.println(" (done)");
// 	    }

// 	    /* Execute main/0 if defined. Look for a single solution. */
// 	    if(pt.db.get("main", 0) != null) {
// 		pt.once(new GoalList(new Compound("main")));
// 	    }
// 	} catch (Exception e) {
// 	    e.printStackTrace();
// 	}

	Botworld botworld = new Botworld();
	BotworldWindow window = new BotworldWindow(botworld);

        window.pack();
 	botworld.setup(db, numBots, world, time);
        window.setVisible(true);
	botworld.start();
    }
}
