package prolog.lib;

import glog.*;

// import java.io.*;

import prolog.*;
import prolog.lib.is_2;
import prolog.math.Numeric;
import prolog.util.FIFOQueue;
import prolog.util.Queue;
import prolog.util.Vector;

/**
 * <TT>gplot/2</TT>: Draw the image specified by the argument. See
 * accompanying docs for more information.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: gplot_1.java,v 1.6 2000/08/15 14:23:51 peteg Exp $
 */
public class gplot_1 extends CompiledPredicate {
    private final static double deg2rad = Math.PI / 180.0;

    public gplot_1() {
    }

    public boolean exec(Term[] args, Prolog p) throws ArgumentException {
 	try {
	    GLog glog = (GLog)p.obj;
	    Compound arg0 = (Compound)args[0].getBinding();
	    glog.clear();
	    draw(arg0, glog);

	    return true;
 	} catch(ClassCastException e) {
	    throw new ArgumentException("gplot/1: invalid argument: ", args);
 	}
    }

    /**
     * Draw the image specified by the prolog description. We do a
     * Breadth-First-Search on the potentially infinite graph, terminating
     * when all pending sub-images are smaller than the resolution of the
     * output device.
     */
    private void draw(Compound arg, GLog glog) throws ArgumentException {
	Queue queue = new FIFOQueue();
	float xres = glog.getWidth(),
	    yres = glog.getHeight();

	queue.enqueue(new GLogContext(xres, yres,
				      0.0,
				      xres / 2f, yres / 2f,
				      false,
				      arg));

// 	BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));

	while(!queue.isEmpty()) {
	    GLogContext cur = (GLogContext)queue.dequeue();

// 	    System.err.println(" drawing: " + cur);

// 	    try {
// 		stdin.readLine();
// 	    } catch(Exception e) {}

	    /* Ignore this operation if we're out of resolution. */
	    if(cur.width < 1 || cur.height < 1)
		continue;

	    Compound node = cur.operation;

	    String cmd = node.functor;
//  	    System.err.println("gplot.draw: cmd = " + cmd);

	    if(cmd == Parser.empty_list_functor) {
		continue;
	    } else if(cmd == Parser.list_functor) {
		while(cmd == Parser.list_functor) {
		    queue.enqueue(new GLogContext(cur, (Compound)node.args[0].getBinding()));
		    node = (Compound)node.args[1].getBinding();
		    cmd = node.functor;
//  		    System.err.println("gplot.draw: list cmd = " + cmd);
		}
		continue;
	    } else if(cmd.equals("above")) {
		above(node, cur, queue);
		continue;
	    } else if(cmd.equals("beside")) {
		beside(node, cur, queue);
		continue;
	    } else if(cmd.equals("flipped")) {
		flip(node, cur, queue);
		continue;
	    } else if(cmd.equals("-")) {
		drawLine(node, glog, cur);
		continue;
	    } else if(cmd.equals("rotated")) {
		rotate(node, cur, queue);
		continue;
	    } else {
		throw new ArgumentException("gplot/1: unknown directive: ", node);
	    }
	}
    }

    void above(Compound node, GLogContext cur, Queue queue) throws ArgumentException {
	Compound uimage = (Compound)node.args[0].getBinding(),
	    bimage = (Compound)node.args[1].getBinding();
	double ratio = 1.0, height = cur.height, t;

	if(bimage.functor.equals("ratio") && bimage.arity == 2) {
	    Term arg1 = bimage.args[1].getBinding();
	    ratio = ((arg1 instanceof Compound) ?
		     is_2.eval((Compound)arg1) : ((Numeric)arg1)).doubleValue();
	    bimage = (Compound)bimage.args[0].getBinding();
	} else if(node.arity != 2) {
	    throw new ArgumentException("gplot.above: invalid above/2: ", node);
	}

	height /= (ratio + 1);

	GLogContext ugc = new GLogContext(cur, uimage);
	ugc.height = ratio * height;
	t = - height / 2.0;

// 	System.err.println(">>> angle = " + cur.angle);
// 	System.err.println(">>> ugc.x = " + ugc.centreX + " ugc.y = " + ugc.centreY);

	ugc.centreX = t * Math.sin(cur.angle) + cur.centreX;
	ugc.centreY = t * Math.cos(cur.angle) + cur.centreY;

// 	System.err.println(">>> ugc.x' = " + ugc.centreX + " ugc.y' = " + ugc.centreY);

	queue.enqueue(ugc);

	GLogContext bgc = new GLogContext(cur, bimage);
	bgc.height = height;
	t = ratio * height / 2.0;

// 	System.err.println(">>> bgc.x = " + bgc.centreX + " bgc.y = " + bgc.centreY);

	bgc.centreX = t * Math.sin(cur.angle) + cur.centreX;
	bgc.centreY = t * Math.cos(cur.angle) + cur.centreY;

// 	System.err.println(">>> bgc.x' = " + bgc.centreX + " bgc.y' = " + bgc.centreY);

	queue.enqueue(bgc);
    }

    void beside(Compound node, GLogContext cur, Queue queue) throws ArgumentException {
	Compound limage = (Compound)node.args[0].getBinding(),
	    rimage = (Compound)node.args[1].getBinding();
	double ratio = 1.0, width = cur.width, t;

	if(rimage.functor.equals("ratio") && rimage.arity == 2) {
	    Term arg1 = rimage.args[1].getBinding();
	    ratio = ((arg1 instanceof Compound) ?
		     is_2.eval((Compound)arg1) : ((Numeric)arg1)).doubleValue();
	    rimage = (Compound)rimage.args[0].getBinding();
	} else if(node.arity != 2) {
	    System.err.println("gplot.beside: invalid beside/2 use.");
	    throw new ArgumentException("gplot.beside: invalid beside/2 use: ", node);
	}

	width = width / (ratio + 1);

// 	System.err.println("beside: ratio = " + ratio + ", width = " + width);

	GLogContext lgc = new GLogContext(cur, limage);
	lgc.width = ratio * width;
	t = - width / 2.0;

// 	System.err.println(">>> angle = " + cur.angle);
// 	System.err.println(">>> lgc.x = " + lgc.centreX + " lgc.y = " + lgc.centreY);

	lgc.centreX = t * Math.cos(cur.angle) + cur.centreX;
	lgc.centreY = t * Math.sin(cur.angle) + cur.centreY;

// 	System.err.println(">>> lgc.x' = " + lgc.centreX + " lgc.y' = " + lgc.centreY);

	queue.enqueue(lgc);

	GLogContext rgc = new GLogContext(cur, rimage);
	rgc.width = width;
	t = ratio * width / 2.0;

// 	System.err.println(">>> rgc.x = " + rgc.centreX + " rgc.y = " + rgc.centreY);

	rgc.centreX = t * Math.cos(cur.angle) + cur.centreX;
	rgc.centreY = t * Math.sin(cur.angle) + cur.centreY;

// 	System.err.println(">>> rgc.x' = " + rgc.centreX + " rgc.y' = " + rgc.centreY);

	queue.enqueue(rgc);
    }

    void flip(Compound node, GLogContext cur, Queue queue) {
	cur.flipped = (cur.flipped) ? false : true;
	cur.operation = (Compound)node.args[0].getBinding();
	queue.enqueue(cur);
    }

    void rotate(Compound node, GLogContext cur, Queue queue) {
	cur.angle = (cur.angle + Math.PI / 2.0) % (2 * Math.PI);
	cur.operation = (Compound)node.args[0].getBinding();
	queue.enqueue(cur);
    }

    private Vector points = new Vector();

    /**
     * Lines look like this: <TT>-(-(-(p1, p2), p3), p4)</TT>, where
     * <I>p<sub>i</sub></I> are points.
     */
    void drawLine(Compound node, GLog glog, GLogContext cur) throws ArgumentException {
	points.removeAllElements();

	// Accumulate the points into a Vector.
	while(node.functor.equals("-") && node.arity == 2) {
	    points.addElement(node.args[1].getBinding());
	    node = (Compound)node.args[0].getBinding();
	}

	// The left-most leaf should be a point as well.
	points.addElement(node);

	int num_points = points.size();
	int[] xs = new int[num_points];
	int[] ys = new int[num_points];

	double width = cur.width, height = cur.height,
	    angle = cur.angle,
	    x0 = cur.centreX, y0 = cur.centreY;
	boolean flipped = cur.flipped;

	for(int i = 0; i < num_points; i++) {
	    Compound thisPoint = (Compound)points.elementAt(i);

	    if(thisPoint.arity != 2 || !thisPoint.functor.equals(",")) {
		throw new ArgumentException("gplot.drawLine: invalid point: ", node);
	    }

	    Term arg0 = thisPoint.args[0].getBinding(),
		arg1 = thisPoint.args[1].getBinding();

	    double x = ((arg0 instanceof Compound) ?
			is_2.eval((Compound)arg0) : ((Numeric)arg0)).doubleValue();
	    double y = ((arg1 instanceof Compound) ?
			is_2.eval((Compound)arg1) : ((Numeric)arg1)).doubleValue();

// 	    System.err.println("drawLine: x = " + x + " y = " + y + ", angle = " + angle + " (" + angle / deg2rad + "), flipped = " + flipped);
// 	    System.err.println("\tx0 = " + x0 + " y0 = " + y0);

	    /* Change the x co-ordinate if flipped around vertical bisector. */
	    x = (flipped) ? 1.0 - x : x;

	    /* Translate (x, y) so that we can rotate around the centre of
               the image. */
	    x -= (1.0 / 2.0);
	    y -= (1.0 / 2.0);

 	    /* Rotate around (0, 0) in translated space. */
 	    double x1 = width * x * Math.cos(angle) - height * y * Math.sin(angle);
 	    double y1 = width * x * Math.sin(angle) + height * y * Math.cos(angle);

// 	    System.err.println("\tx1 = " + (x0 + x1) + " y1 = " + (y0 + y1));

	    /* Translate to absolute co-ordinates. */
	    xs[i] = (int)(x0 + x1 - 0.5);
	    ys[i] = (int)(y0 + y1 - 0.5);
	}

	glog.drawLine(xs, ys);
    }
}
