package prolog;

import java.io.*;

/**
 * A do-nothing-but-the-minimum clause index.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: ClauseIndex.java,v 1.19 2000/11/07 12:04:55 peteg Exp $
 */
class ClauseIndex implements Cloneable, Serializable {
    /**
     * The list of clauses that form this predicate.
     */
    protected ClauseIndexRec cir;

    ClauseIndex() {
	this.cir = new ClauseIndexRec();
    }

    /**
     * Do a goal, i.e. the first time.
     */
    public boolean first(Term[] args, Prolog p) throws PrologException {
//  	System.err.println("ClauseIndex.unify(): goal = " + goal);

	return unify(args, p, getClauses(args));
    }

    /**
     * Redo a goal, i.e. on backtracking.
     */
    public boolean redo(Term[] args, Prolog p, ChoicePoint cp) throws PrologException {
//   	System.err.println("ClauseIndex.redo(): goal = " + goal
//  			   + ", next_clause: " + cp.next_clause);

	return unify(args, p, cp.next_clause);
    }

    /**
     * Unify a goal against the clauses in this predicate.
     */
    protected boolean unify(Term[] args, Prolog p, ClauseList cl) {
	Binding trail_top = p.get_trail_top();

	for(; cl != null; cl = cl.next) {
	    Clause clause = cl.clause;
	    Compound head = clause.head;
	    Term[] insts = new Term[clause.numVars];

	    if(head.unify_args(args, insts, p)) {
//   		System.err.println("ClauseIndex.unify(): success with clause:\n"
// 				   + cl + "with insts:");
// 		for(int i = 0; i < insts.length; i++) {
// 		    System.err.println(i + " -> " + insts[i] + "\n");
// 		}
		
		if(cl.next != null) {
		    /* Have alternative, so allocate a ChoicePoint. */
		    p.allocate_cp(cl.next, trail_top);
		}

		/* Allocate an ActivationRecord if necessary. */
		p.allocate_ar(insts, clause.body);

		return true;
	    } else {
		/* Doesn't match. Unwind the trail, try the next one. */
//  		System.err.println("ClauseIndex.unify(): clause doesn't match: " + clause + " <-> ");

// 		for(int i = 0; i < args.length; i++) {
// 		    System.err.println(i + " -> " + args[i]);
// 		}

		p.unwind_trail(trail_top);
	    }
	}

	return false;
    }

    /**
     * Performs no indexing - returns all clauses in this predicate.
     *
     * @return An ordered list of all clauses in this predicate.
     */
    protected final ClauseList getClauses() {
	return cir.firstClause;
    }

    /**
     * Gets a list of clauses that could unify against the given term. This
     * method can be overridden by indices that filter the clause list,
     * rather than combine indexing with unification.
     *
     * @see prolog.ClauseIndexTrad
     */
    protected ClauseList getClauses(Term[] args) {
	return cir.firstClause;
    }

    /**
     * Adds a clause to the beginning of the predicate.
     */
    protected void puta(Clause c) {
	cir.puta(c);
    }

    /**
     * Adds a clause to the end of the predicate.
     */
    protected void putz(Clause c) {
	cir.putz(c);
    }

    /**
     * Removes all clauses from this index.
     */
    public void clear() {
	cir.clear();
    }

    /**
     * Removes a clause from this index.
     */
    void remove(Clause c) {
	cir.remove(c);
    }

    /**
     * Compile each clause that could match the given type.
     */
    void compileClauses(PredicateCodeGen pcg, TermInfo[] arg_tis) {
	for(ClauseList cl = cir.firstClause; cl != null; cl = cl.next) {
	    pcg.compileClause(cl.clause, cl.next == null);
	}

	// If none of the clauses succeeded, we fail.
	pcg.failure();
    }

    public String toString() {
	return "ClauseIndex:\n" + cir.toString();
    }
}
