package prolog;

import java.io.*;

import prolog.lib.*;

import prolog.util.Hashtable;
import prolog.util.Strings;
import prolog.util.Vector;

/**
 * Represents <TT>Compound</TT> terms (of the form <TT>functor(arg1, arg2,
 * ... )</TT>).
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Compound.java,v 1.33 2000/11/16 10:30:07 peteg Exp $
 */
public class Compound extends Term implements Cloneable {
    /**
     * A unique value that represents this <TT>functor/arity</TT>
     * combination.
     */
    public FunctorArity fa;
    public int functor_arity;

    public String functor;
    public int arity;
    public Term[] args;

    /**
     * Indicates that this <TT>Compound</TT> doesn't need to be copied
     * (<TT>true</TT>, e.g. if this is a ground term, or all
     * <TT>Variable</TT>s have been replaced with <TT>Binding</TT>s, or
     * <TT>arity == 0</TT>), or that it might need to be (<TT>false</TT>).
     */
    boolean copied;

    /**
     * Creates a <TT>Compound</TT> with given functor and arguments.
     *
     * @param f the functor of the <TT>Term</TT>
     * @param arglist an array of arguments
     */
    public Compound(String functor, Term[] arglist) {
	this.functor = functor.intern();
	this.arity   = (arglist == null) ? 0 : arglist.length;
	this.args    = arglist;

	this.fa            = map_functor_arity(functor, arity);
	this.functor_arity = fa.functor_arity;

	this.copied = (arity == 0) ? true : false;

//  	System.err.println("new Compound(): " + this);
    }

    /**
     * Creates a <TT>Compound</TT> with given functor and zero arity. Used
     * by facts and strings.
     *
     * @param f the functor of the <TT>Term</TT>
     */
    public Compound(String functor) {
	this(functor, null);
	copied = true;
    }

    /**
     * Creates a <TT>Compound</TT> with given functor and arguments.
     *
     * @param f the functor of the <TT>Term</TT>
     * @param arglist an array of arguments
     * @see prolog.ClauseCodeGen#compound_create
     */
    public Compound(FunctorArity fa, Term[] args) {
	this.functor = fa.functor;
	this.arity   = fa.arity;
	this.args    = args;

	this.fa            = fa;
	this.functor_arity = fa.functor_arity;

	this.copied = true;

//   	System.err.println("new Compound(): " + this);
    }

    /**
     * Creates a <TT>Compound</TT> with given functor and zero arity. Used
     * by facts and strings.
     *
     * @param f the functor of the <TT>Term</TT>
     */
    public Compound(FunctorArity fa) {
	this.functor = fa.functor;
	this.arity   = fa.arity;

	this.fa            = fa;
	this.functor_arity = fa.functor_arity;

	this.copied = true;

//   	System.err.println("new Compound(): " + this);
    }

    /**
     * Don't bother checking the functor.
     */
    public Term[] instantiateHead(Term[] insts) {
	if(copied) {
	    return args;
	} else {
	    Term[] new_args = new Term[arity];

	    for(int i = arity; i-- > 0; ) {
// 		System.err.println("Compound.instantiateHead(" + this + "): "
// 				   + i + " -> " + args[i]);
		new_args[i] = args[i].instantiate(insts);
	    }

	    return new_args;
	}
    }

    public Term instantiate(Term[] insts) {
//  	System.err.println("Compound.instantiate(): " + this);

	if(copied) {
// 	    System.err.println(" (do nothing)");
	    return this;
	} else {
// 	    System.err.println(" (instantiating)");

	    Term[] new_args = new Term[arity];

	    for(int i = arity; i-- > 0; ) {
// 		    c.args[i] = (args[i] == null) ? null : args[i].instantiate(insts);
		new_args[i] = args[i].instantiate(insts);
	    }

	    return new Compound(fa, new_args);
	}
    }

    /**
     * Unifies this <TT>Compound</TT> term against a list of arguments,
     * assuming that the functor and arity match.
     *
     * @see prolog.InterpretedClause#unify
     */
    public boolean unify_args(Term[] bs, Term[] insts, Prolog p) {
//    	System.err.println("Compound.unify_args(): " + this);

	for(int i = 0; i < arity; i++) {
// 	    System.err.println(" >>   args[" + i + "] = " + args[i]);
// 	    System.err.println(" >> c.args[" + i + "] = " + bs[i]);

	    // This stuff is for compiled code which might have null
	    // ref's == uninstantiated vars.
	    Term c_arg = bs[i];

// 	    if(args[i] == null) {
// 		args[i] = c_arg;
// 	    } else if(c_arg == null) {
// 		bs[i] = args[i];
// 	    } else

	    if(!args[i].unify(c_arg, insts, p)) {
		return false;
	    }
	}

// 	System.err.println(" >> success.");

	return true;
    }

    public boolean unify(Term t, Term[] insts, Prolog p) {
//    	System.err.println("Trying to unify: " + this + " with: " + t);

	if(t instanceof Variable) {
	    return t.unify(this, insts, p);
	}
	
	if(!(t instanceof Compound)) {
	    return false;
	}

	Compound c = (Compound)t;

	if(functor_arity == c.functor_arity) {
	    return unify_args(c.args, insts, p);
	} else {
// 	    System.err.println("unify failed: " + this + " <-> " + c);
	    return false;
	}
    }

    public boolean identical(Term t) {
	if(!(t instanceof Compound)) {
	    return false;
	}

	Compound c = (Compound)t;

	if(functor_arity == c.functor_arity) {
	    if(arity > 0) {
		for(int i = arity; i-- > 0; ) {
		    if(!args[i].identical(c.args[i])) {
			return false;
		    }
		}
	    }

	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Squash all bound variables (to their bindings) and substitute
     * <TT>Variable</TT>s for <TT>Binding</TT>s. Used by the
     * <TT>assert/1</TT>s so that the database only contains ground terms
     * and unbound variables. This isn't pretty.
     *
     * @see prolog.lib.asserta_1
     * @see prolog.lib.assertz_1
     */
    protected Compound squash(Hashtable vars) {
 	Variable v;

	if(arity == 0) {
	    return this;
	} else {
	    try {
		Compound c = (Compound)this.clone();
		c.copied = (arity == 0) ? true : false;

		Term[] new_args = c.args = new Term[arity];

		for(int i = arity; i-- > 0; ) {
		    new_args[i] = args[i].getBinding();

		    if(new_args[i] instanceof Compound) {
			new_args[i] = ((Compound)new_args[i]).squash(vars);
		    } else {
			/* If it's a Binding, change it into a Variable. */
			if(new_args[i] instanceof Binding) {
			    if((v = (Variable)vars.get(new_args[i])) == null) {
				vars.put(new_args[i],
					 (v = new Variable(vars.size())));
			    }

			    new_args[i] = v;
			}
		    }
		}

		return c;
	    } catch(CloneNotSupportedException e) {
		throw new InternalError(e.toString());
	    }
	}
    }


    public Term normalise_copy(Variable[] vars,
			       int next_new_var,
			       Vector new_vars) {
	Term[] new_args = new Term[arity];

	for(int i = 0; i < arity; i++) {
	    int old_new_vars_size = new_vars.size();

	    new_args[i] = args[i].normalise_copy(vars, next_new_var, new_vars);

	    next_new_var += new_vars.size() - old_new_vars_size;
	}

	return new Compound(fa, new_args);
    }

    public void typify(TypeDictionary types, boolean map) {
// 	if(ti == null || ti.contains_variables) {
//  	if(ti == null) {
// 	    System.err.println("COMPOUND.TYPEIFY()");

	    types.type_of(this, map);
//  	}
    }



    /**
     * Is this <TT>Compound</TT> term (thought of as a predicate) a
     * unification?
     */
    boolean isUnify() {
	return functor == unify_2.functor && arity == 2;
    }

    /**
     * Is this goal an <TT>is/2</TT> invocation?
     */
    boolean isIs() {
	// FIXME
	return arity == 2 && (
			      functor == is_2.functor
		           || functor == lt_2.functor
			   || functor == is_not_equal_2.functor
			      );
    }

    public String displayString() {
	StringBuffer out = new StringBuffer();

	/* Pretty-print a list. */
	if(functor == Parser.list_functor) {
	    out.append("[");

	    Compound cur = this;

	    while(true) {
		Term t0 = cur.args[0];
		Term t1 = cur.args[1].getBinding();

		out.append(t0.displayString());

		if(t1 instanceof Compound) {
		    Compound next = (Compound)t1;

		    if(next.functor == Parser.list_functor && next.arity == 2) {
			out.append(", ");
			cur = next;
			continue;
		    } else if(next.functor == Parser.empty_list_functor) {
			break;
		    }
		}

		out.append("|");
		out.append(t1.displayString());
		break;
	    }

	    out.append("]");
	} else {
	    out.append(functor);

	    if(arity > 0) {
		out.append("(");

		for(int i = 0; i < arity; ) {
		    Term arg = args[i];

		    out.append(arg == null ? " < null > " : arg.displayString());

		    if(++i < arity) {
			out.append(", ");
		    }
		}
		out.append(")");
	    }
	}

	return out.toString();
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	if(functor == Parser.list_functor) {
	    out.append("[");

	    Compound cur = this;

	    while(true) {
		Term t0 = cur.args[0];
		Term t1 = cur.args[1];

		out.append((t0 == null) ? null : t0.toString());

		if(t1 != null) {
		    t1 = t1.getBinding();

		    if(t1 instanceof Compound) {
			Compound next = (Compound)t1;

			if(next.functor == Parser.list_functor && next.arity == 2) {
			    out.append(", ");
			    cur = next;
			    continue;
			} else if(next.functor == Parser.empty_list_functor) {
			    break;
			}
		    }
		}

		out.append("|");
		out.append((t1 == null) ? null : t1.toString());
		break;
	    }

	    out.append("]");
	} else {
	    String f = Strings.escape_string(functor);
	
	    /* If escaping did anything, enclose it in parenthesis. */
	    if(f == functor) {
		out.append(f);
	    } else {
		out.append("\"");
		out.append(f);
		out.append("\"");
	    }

	    if(arity > 0) {
		out.append("(");

		for(int i = 0; i < arity; ) {
		    out.append((args[i] == null) ? null : args[i].toString());
		    if(++i < arity) {
			out.append(", ");
		    }
		}
		out.append(")");
	    }
	}

	return out.toString();
    }

    /**
     * Object serialisation breaks the String.intern() method, which we rely
     * on to speed up functor equality checks (pointer equality v's string
     * equality). So we fix that up here...
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException {
	in.defaultReadObject();
	functor = functor.intern();
    }

    private static Hashtable functor_arity_hash = new Hashtable();

    public static synchronized FunctorArity map_functor_arity(String functor, int arity) {
	FunctorArity fa;
	String key = functor + "/" + arity;

	if((fa = (FunctorArity)functor_arity_hash.get(key)) == null) {
	    functor_arity_hash.put(key,
				   (fa = new FunctorArity(functor, arity)));
	}

	return fa;
    }
}
