package prolog;

import prolog.util.Hashtable;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * Represents <TT>Compound</TT> term types.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: CompoundType.java,v 1.23 2000/11/16 10:30:07 peteg Exp $
 */
public class CompoundType extends TermType {
    final static String ClassName    = "prolog.Compound";
    final static ObjectType JavaType = new ObjectType("prolog/Compound");

    public FunctorArity fa;
    int functor_arity;

    public String functor;
    public int arity;
    public TermType[] args;

    /**
     * Constructor for use by the <TT>TypeDictionary</TT> (typedefs).
     *
     * @see prolog.TypeDictionary
     */
    CompoundType(Compound c, TypeDictionary types, Hashtable vars)
	throws ArgumentException {

	this.fa            = c.fa;
	this.functor_arity = c.functor_arity;

	this.functor = c.functor;
	this.arity   = (c.args == null) ? 0 : c.args.length;
	this.args    = new TermType[arity];

	for(int i = 0; i < arity; i++) {
	    Term t = c.args[i];

	    if(t instanceof Binding) {
		this.args[i] = (TypeVariable)vars.get(t);
// 		System.err.println("  " + t + " -> " + args[i]);
	    } else {
		Compound arg = (Compound)t;
		TermType arg_t;

		if((arg_t = types.get(arg.functor_arity)) == null) {
		    arg_t = new CompoundType((Compound)t, types, vars);
		}

		this.args[i] = arg_t;
	    }

	    if(args[i] == null) {
		throw new ArgumentException("TypeDictionary.type_def(): unknown type definition element: " + t);
	    }
	}
    }

    /**
     * Constructs a new <TT>CompoundType</TT> object, with all args fully
     * specified.
     *
     * @param map map functors to their most general type.
     * @see prolog.CompoundType#make
     */
    private CompoundType(Compound c, TermType[] args) {
	this.fa            = c.fa;
	this.functor_arity = c.functor_arity;

	this.functor       = c.functor;
	this.arity         = args.length;
	this.args          = args;
    }

    /**
     * Wrapper to build TI's for <TT>Compound</TT> terms with unspecified
     * (== not <TT>typedef</TT>ed) types.
     *
     * @see prolog.TypeDictionary#typeof
     */
    static TermInfo make(Compound c, TypeDictionary types, boolean map) {
	TermType[] args   = new TermType[c.arity];
	CompoundType type = new CompoundType(c, args);

	// Create a new TI for this term.
	TermInfo[] arg_tis = new TermInfo[c.arity];
	TermInfo ti = c.ti = new TermInfo(type,
					  null,
					  arg_tis);

	for(int i = 0; i < c.arity; i++) {
	    c.args[i].typify(types, map);
	    arg_tis[i] = c.args[i].ti;
	    args[i] = arg_tis[i].type;
	}

	return ti;
    }

    ObjectType getJavaType(TermInfo ti) {
	return JavaType;
    }

    TermInfo instantiate(Term t, TermInfo[] typevars,
			 TypeDictionary types, boolean map) {
	Compound c = (Compound)t;

// 	System.err.println("CompoundType.instantiate(): " + this + " against " + c);

// 	System.err.println(" >> typevars: " + typevars);

	if(functor_arity != c.functor_arity) {
	    throw new RuntimeException("Functor/Arity doesn't match!");
	}

	// Create a new TI for this term.
	TermInfo[] arg_tis = new TermInfo[arity];
	TermInfo ti = c.ti = new TermInfo(this,
					  typevars,
					  arg_tis);

	// Assume there's no variables in this term.
// 	ti.contains_variables = false;

	for(int i = 0; i < arity; i++) {
	    arg_tis[i] = args[i].instantiate(c.args[i], typevars, types, map);

	    // If any subterm has a variable, we do too.
// 	    ti.contains_variables |= arg_tis[i].contains_variables;
	}

	return ti;
    }

    boolean unify(TermInfo ti,
		  TermInfo arg,
		  TypeTrail trail) {
 	System.err.println("CompoundType.unify():\n"
			   + this + " " + ti
			   + "\n<->\n "
			   + arg.type + " " + arg);

	if(arg.type instanceof TypeVariable) {
	    return arg.type.unify(arg, ti, trail);
	}

	if(!(arg.type instanceof CompoundType)) {
	    return false;
	}

	CompoundType ct = (CompoundType)arg.type;

	if(ct.functor_arity != functor_arity) {
	    return false;
	}

	for(int i = 0; i < arity; i++) {
	    TermInfo arg_ti = ti.arg_tis[i];

	    if(!arg_ti.type.unify(arg_ti, arg.arg_tis[i], trail)) {
		return false;
	    }
	}

	return true;
    }

    public boolean compile_unify(ClauseCodeGen ccg,
				 TermInfo arg0, TermInfo arg1) {
//   	System.err.println("CompoundType.compile_unify(): unify: "
// 			   + this
// 			   + " <-> "
// 			   + arg1.type);

 	if(arg1.type instanceof TypeVariable) {
 	    // Reverse data flow.
 	    return arg1.type.compile_unify(ccg, arg1, arg0);
 	}

	// Ensure we've got CompoundTypes on both arg's.
	if(arg1.type instanceof AlgebraicType) {
	    throw new RuntimeException();
// 	    return arg1.type.compile_unify(ccg, arg1, arg0);
	}
	
	if(!(arg1.type instanceof CompoundType)) {
// 	    System.err.println("  >> types don't match.");
	    return false;
	}

	if(((CompoundType)arg1.type).functor_arity != functor_arity) {
// 	    System.err.println("  >> functor/arity doesn't match.");
	    return false;
	}

	// OK, actually compile the damn thing.
	return ccg.compound_unify(arg0, arg1);
    }

    public void create(ClauseCodeGen ccg, TermInfo arg0) {
  	ccg.compound_create(arg0);
    }

    public void toString(StringBuffer out) {
	out.append("t_");
 	out.append(functor);
	out.append("(");

	for(int i = 0; i < arity; ) {
	    out.append(args[i].toString());
	    if(++i < arity) {
		out.append(", ");
	    }
	}

	out.append(")");
    }
}
