package prolog;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

import java.util.Enumeration;

import prolog.util.Hashtable;

/**
 * Represents a pending compilation job.
 * <P>
 * Also provides the structure for tying together <TT>ClauseCode</TT>
 * objects into a <TT>unify()</TT> method.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: PredicateCodeGen.java,v 1.31 2000/11/13 11:55:04 peteg Exp $
 * @see prolog.ClauseCodeGen
 * @see prolog.JITClassLoader
 * @see prolog.JITPredicate
 */
public class PredicateCodeGen {
    /**
     * The ClassLoader. Loads the compiled class definitions into the JVM.
     */
    static final JITClassLoader classloader = new JITClassLoader();

    Prolog prolog;
    TypeDictionary types;
    private JITPredicate predicate;
    private PredicateTypeRec ptr;
    private TermInfo[] type;

    String functor;
    int arity;

    String class_name;

    ClassGen class_gen;
    ConstantPoolGen cp;
    InstructionFactory inst_fact;

    InstructionList il;

    Hashtable predicates;
    Hashtable functors;

    /**
     * Does this predicate require a <TT>ChoicePoint</TT>?
     */
    private boolean needsChoicePoint;

    long startTime;

    public PredicateCodeGen(Prolog prolog,
			    JITPredicate predicate,
			    PredicateTypeRec ptr) {

	this.startTime = System.currentTimeMillis();

	this.prolog    = prolog;
	this.types     = prolog.types;
	this.predicate = predicate;
	this.ptr       = ptr;
 	this.type      = ptr.arg_tis;

	this.il        = new InstructionList();

	this.functor   = predicate.functor;
	this.arity     = predicate.arity;

	this.predicates = new Hashtable();
	this.functors   = new Hashtable();

	StringBuffer cn = new StringBuffer();

	cn.append(JITDefs.package_name);
	cn.append(".");
	cn.append(functor);
	cn.append("_");
	cn.append(arity);
// 	cn.append("_");

	class_name = cn.toString();
    }

    /**
     * Compiles a given predicate (= list of normalised clauses) for the
     * given type signature.
     */
    public void run() {
	/* Setup the JavaClass stuff. */
	this.class_gen = new ClassGen(class_name,
				      JITDefs.extends_class_name,
				      "JIT",
				      JITDefs.ACC_PUBLIC | JITDefs.ACC_FINAL | JITDefs.ACC_SUPER,
				      null);
	this.cp = class_gen.getConstantPool();
	this.inst_fact = new InstructionFactory(class_gen);



// 	System.err.println(">> PredicateCodeGen.compile: name: " + class_name
// 			   + " predicate:\n" + predicate);
// 	System.err.println(">> PredicateCodeGen.compile: type: ");

// 	for(int i = 0; i < type.length; i++) {
// 	    System.err.println("   " + type[i]);
// 	}


	/* Compile each clause in turn -> this.il. */
 	predicate.index.compileClauses(this, type);


	mkUnify();
	mkConstructor();


	/* Load the Class, create an instance. */
 	try {
	    JavaClass clazz = class_gen.getJavaClass();

	    /* Dump to a file for testing purposes. */
//   	    clazz.dump(class_gen.getClassName() + ".class");

	    /* Do the business. */
  	    Class new_pred = classloader.registerClass(clazz);

	    /* Patch the predicate call static fields. */
	    for(Enumeration e = predicates.keys(); e.hasMoreElements(); ) {
		String field_name = (String)e.nextElement();
		Predicate predicate = (Predicate)predicates.get(field_name);

		new_pred.getField(field_name).set(null, predicate);
	    }

	    /* Patch the FunctorArity static fields. */
	    for(Enumeration e = functors.keys(); e.hasMoreElements(); ) {
		String field_name = (String)e.nextElement();
		FunctorArity fa = (FunctorArity)functors.get(field_name);

// 		System.err.println(">> Adding data for field: " + field_name);

		if(fa.arity == 0) {
		    // Whack in a new compound instead.
// 		    System.err.println(" new compound: " + fa);
		    new_pred.getField(field_name).set(null, new Compound(fa));
		} else {
// 		    System.err.println(" just the f/a: " + fa);
		    new_pred.getField(field_name).set(null, fa);
		}
	    }

    	    CompiledPredicate new_pred_inst = (CompiledPredicate)new_pred.newInstance();
  	    ptr.setCompiled(new_pred_inst);

	    // Benchmarking
	    long queryTime = System.currentTimeMillis() - startTime;
	    System.err.println("Compiler Benchmark: " + queryTime);
 	} catch(Throwable e) {
 	    throw new RuntimeException("PredicateCodeGen: output problems: " + e);
 	}
    }

    /**
     * Compiles a particular clause. A call back from the indexer.
     *
     * @param c the clause to compile.
     * @param ci_det is this clause deterministic according to the clause
     * indexer?
     */
    ClauseCodeGen compileClause(Clause c, boolean ci_det) {
// 	System.err.println(">> PredicateCodeGen.compileClause:\n" + c);
	ClauseCodeGen ccg = new ClauseCodeGen(this);

	if(ccg.compile(c, type, ci_det)) {
// 	    System.err.println(">> PredicateCodeGen.compile: compilation succeeded: " + ccg);
	    return ccg;
	} else {
//  	    System.err.println(">> PredicateCodeGen.compile: compilation failed for clause:\n" + c);
	    return null;
	}
    }

    /**
     * Generates the code for the <TT>unify()</TT> method.
     *
     * @see prolog.CompiledClause
     */
    private void mkUnify() {
	// So we can give each instruction a "line number".
	il.setPositions(false);

	MethodGen unify = new MethodGen(JITDefs.ACC_PUBLIC,
					JITDefs.unify_ret_type,
					JITDefs.unify_arg_type,
					JITDefs.unify_arg_names,
					JITDefs.unify_name,
					class_name,
					il,
					cp);

	unify.addException(JITDefs.prologException_name);
//   	unify.removeNOPs();
 	unify.setMaxStack(JITDefs.unify_stack_depth);
	unify.setMaxLocals();

	/* Add the static fields for called predicate records. */
	for(Enumeration e = predicates.keys(); e.hasMoreElements(); ) {
	    String field_name = (String)e.nextElement();

// 	    System.err.println(" >> new field for: " + field_name);
	    class_gen.addField(new Field(JITDefs.ACC_PUBLIC | JITDefs.ACC_STATIC,
					 cp.addUtf8(field_name),
 					 cp.addUtf8(JITDefs.predicate_type.getSignature()),
					 null,
					 cp.getConstantPool()));
	}

	/* Add the static fields for called FunctorArity's. */
	for(Enumeration e = functors.keys(); e.hasMoreElements(); ) {
	    String field_name = (String)e.nextElement();

// 	    System.err.println(" >> new field for: " + field_name);

	    if(field_name.charAt(0) == 'c') {
// 		System.err.println("  for a compound.");
		class_gen.addField(new Field(JITDefs.ACC_PUBLIC | JITDefs.ACC_STATIC,
					     cp.addUtf8(field_name),
					     cp.addUtf8(CompoundType.JavaType.getSignature()),
					     null,
					     cp.getConstantPool()));
	    } else {
// 		System.err.println("  for a f/a.");
		class_gen.addField(new Field(JITDefs.ACC_PUBLIC | JITDefs.ACC_STATIC,
					     cp.addUtf8(field_name),
					     cp.addUtf8(FunctorArity.JavaType.getSignature()),
					     null,
					     cp.getConstantPool()));
	    }
	}

	/* Add line number info to the unify() method. The ith line
	 * is the ith instruction. */
//  	InstructionHandle[] ihs = il.getInstructionHandles();

//  	for(int i = 0; i < ihs.length; i++) {
// 	    unify.addLineNumber(ihs[i], ihs[i].getPosition());
// 	}

   	class_gen.addMethod(unify.getMethod());	

// 	System.err.println(">> PredicateCodeGen.mkUnify(): generated code:\n" + il);
    }


    /**
     * Append "return failure" code.
     */
    public InstructionHandle failure() {
	InstructionHandle fail = il.append(inst_fact.ICONST_0);
	il.append(inst_fact.IRETURN);

	return fail;
    }



    /**
     * Init code to handle backtracking into this predicate.
     *
     * Stack: <I> ... -> ...</I>
     */
//     public void choice_point_init(InstructionList il) {
// 	System.err.println("choice_point_init()");
	
// 	il.append(inst_fact.ICONST_M1);
// 	il.append(inst_fact.createInvoke(JITDefs.prolog_class_name,
// 					 "setCN",
// 					 Type.VOID,
// 					 new Type[]{Type.INT},
// 					 JITDefs.INVOKEVIRTUAL));
//     }

//     /**
//      * Stack: <I> ... -> ...</I>
//      */
//     public void choice_point_set(InstructionList il, int clauseNum) {
// 	System.err.println("IS.choice_point_set(): " + clauseNum);
// 	il.append(inst_fact.NOP);
// 	il.append(inst_fact.NOP);
// 	il.append(inst_fact.NOP);
// 	il.append(inst_fact.NOP);
// 	il.append(JITDefs.prolog_load_inst);
// 	il.append(JITDefs.prolog_load_inst);

// 	il.append(JITDefs.prolog_load_inst);
// 	il.append(new PUSH(cp, clauseNum));
// 	il.append(inst_fact.createInvoke(JITDefs.prolog_class_name,
// 					 "setCN",
// 					 Type.VOID,
// 					 new Type[]{Type.INT},
// 					 JITDefs.INVOKEVIRTUAL));
//     }

//     /**
//      * Stack: <I> ... -> ...</I>
//      */
//     public void choice_point_finished(InstructionList il) {
// 	System.err.println("IS.choice_point_finished()");
// 	il.append(JITDefs.prolog_load_inst);
// 	il.append(inst_fact.ICONST_M1);
// 	il.append(inst_fact.createInvoke(JITDefs.prolog_class_name,
// 					 "setCN",
// 					 Type.VOID,
// 					 new Type[]{Type.INT},
// 					 JITDefs.INVOKEVIRTUAL));
//     }

    /**
     * Builds a constructor for the compiled predicate.
     *
     * @see prolog.CompiledClause
     */
    private void mkConstructor() {
	InstructionList constil = new InstructionList();
	MethodGen constructor;

	/* Invoke super's constructor, with "do we need a choicepoint?" as argument. */
	constil.append(inst_fact.createThis());
	constil.append(inst_fact.createInvoke(JITDefs.extends_class_name,
					      JITDefs.pred_init_name,
					      JITDefs.pred_init_ret_type,
					      JITDefs.pred_init_arg_type,
					      JITDefs.INVOKESPECIAL));

	constil.append(inst_fact.RETURN);

	/* Build the method */
	constructor = new MethodGen(JITDefs.ACC_PUBLIC,
				    Type.VOID,
				    Type.NO_ARGS,
				    null,
				    JITDefs.pred_init_name,
				    class_name,
				    constil,
				    cp);
 	constructor.setMaxStack(1);

	class_gen.addMethod(constructor.getMethod());
    }

    public String toString() {
	return ">> PredicateCodeGen: needsChoicePoint? " + needsChoicePoint;
    }
}
