package prolog.math;

import prolog.*;
import prolog.util.Vector;

/**
 * Maths package for Prolog. A stripped and modified version of the gnu.math
 * library, which has the following copyright:
 * <P>
 * Copyright (c) 1997  Per M.A. Bothner.
 * <B>This is free software;  for terms and warranty disclaimer see ./COPYING.</B>
 * <P>
 * A class for infinite-precision integers.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: IntNum.java,v 1.12 2000/11/13 11:55:05 peteg Exp $
 */
public class IntNum extends Numeric {
    public int ival;

    public Term normalise_copy(Variable[] vars,
			       int next_new_var,
			       Vector new_vars) {
	return new IntNum(ival);
    }

    public void typify(TypeDictionary types, boolean map) {
	if(ti == null) {
	    ti = new TermInfo(new IntType(this));
	}

// 	System.err.println(">> IntNum.typeify(): " + this
// 			   + " -> .ti: " + ti);
    }

    /**
     * We pre-allocate integers in the range minFixNum..maxFixNum.
     */
    static final int minFixNum = -100;
    static final int maxFixNum = 1024;
    static final int numFixNum = maxFixNum - minFixNum + 1;
    static final IntNum[] smallFixNums = new IntNum[numFixNum];

    static {
	for(int i = numFixNum; --i >= 0; ) {
	    smallFixNums[i] = new IntNum(i + minFixNum);
	}
    }

    public IntNum() {
    }

    /**
     * Create a new (non-shared) <TT>IntNum</TT>, and initialize to an
     * <TT>int</TT>.
     *
     * @param value the initial value
     */
    public IntNum(int value) {
	ival = value;
    }

    /**
     * Return a (possibly-shared) IntNum with a given int value.
     */
    public static IntNum make(int value) {
	if(value >= minFixNum && value <= maxFixNum) {
	    return smallFixNums[(int)value - minFixNum];
	} else {
	    return new IntNum(value);
	}
    }

    public Numeric add(Numeric y) {
	if(y instanceof IntNum) {
	    return IntNum.make(ival + ((IntNum)y).ival);
	} else {
	    // y instanceof FloNum
	    return new FloNum(ival + ((FloNum)y).value);
	}
    }

    public IntNum add(IntNum y) {
	return IntNum.make(ival + y.ival);
    }

    public Numeric sub(Numeric y) {
	if(y instanceof IntNum) {
	    return IntNum.make(ival - ((IntNum)y).ival);
	} else {
	    // y instanceof FloNum
	    return new FloNum(ival - ((FloNum)y).value);
	}
    }

    public IntNum sub(IntNum y) {
	return IntNum.make(ival - y.ival);
    }

    public Numeric mul(Numeric y) {
	if (y instanceof IntNum) {
	    return IntNum.make(ival * ((IntNum)y).ival);
	} else {
	    // y instanceof FloNum
	    return new FloNum(ival * ((FloNum)y).value);
	}
    }

    public Numeric div(Numeric y) {
	double yval;

	if(y instanceof IntNum) {
	    yval = ((IntNum)y).ival;
	} else {
	    // y instanceof FloNum
	    yval = ((FloNum)y).value;
	}

	return new FloNum((double)ival / yval);
    }

    public Numeric mod(Numeric y) {
// 	System.err.print("IntNum: %: x = " + this + " y = " + y);

	if(y instanceof IntNum) {
// 	    System.err.println(" (int)  = " + IntNum.modulo(this, (IntNum)y));
	    return IntNum.make(ival % ((IntNum)y).ival);
	} else {
	    // y instanceof FloNum
	    return new FloNum(ival % ((FloNum)y).value);
	}
    }

    public Numeric neg() {
	return IntNum.make(-ival);
    }

    public static final IntNum one() {
	return smallFixNums[1 - minFixNum];
    }

  /** Return the IntNum for -1. */
//   public static IntNum minusOne ()
//   {
//     return smallFixNums[-1 - minFixNum];
//   }

    public final boolean isNegative() {
	return ival < 0;
    }

    public int sign() {
	return ival > 0 ? 1 : ival < 0 ? -1 : 0;
    }

    public int compare(Numeric n) {
	if(n instanceof IntNum) {
	    int y = ((IntNum)n).ival;

// 	    System.err.println("IntNum.compare(): " + this + " <-> " + y
// 			       + " -> " + (ival < y ? -1 : ival > y ? 1 : 0));

	    return ival < y ? -1 : ival > y ? 1 : 0;
	} else {
	    // n instanceof FloNum
	    double y = ((FloNum)n).value;

	    return ival < y ? -1 : ival > y ? 1 : 0;
	}
    }


//     public final boolean isZero() {
//   {
//     return words == null && ival == 0;
//   }

//   public final boolean isOne ()
//   {
//     return words == null && ival == 1;
//   }

//   public final boolean isMinusOne ()
//   {
//     return words == null && ival == -1;
//   }


    public Numeric toInt(int rounding_mode) {
	return this;
    }

    public String toString(int radix) {
	return Integer.toString(ival, radix);
    }

    public int intValue() {
	return ival;
    }

    public long longValue() {
	return ival;
    }

    public int hashCode() {
	return ival;
    }

    public boolean equals(Object obj) {
	if (obj == null || ! (obj instanceof IntNum)) {
	    return false;
	}

	return ival == ((IntNum)obj).ival;
    }

    public double doubleValue() {
	return (double)ival;
    }
}
