% -*- Prolog -*-

% Computational Intelligence: a logical approach. 
% Prolog Code. 
% A CSP solver using arc consistency (Figure 4.8)
% Copyright (c) 1998, Poole, Mackworth, Goebel and Oxford University Press.
                                    
% csp(Domains, Relations) means that each variable has
% an instantiation to one of the values in its Domain 
% such that all the Relations are satisfied.
% Domains represented as list of 
% [dom(V,[c1,...,cn]),...]
% Relations represented as [rel([X,Y],r(X,Y)),...]
%  for some r
csp(Doms, Relns) :-
	ac(Doms, Relns).

% ac(Dom,Relns) is true if the domain constrants
% specified in Dom and the binary relations
% constraints specified in Relns are satisfied.
ac(Doms, Relns) :-
	make_arcs(Relns, A),
%	write(Relns), nl, write(A), nl,
	consistent(Doms, [], A, A).

% make_arcs(Relns, Arcs) makes arcs Arcs corresponding to
% relations Relns. There are acrs for each ordering of
% variables in a relations.
make_arcs([], []).
make_arcs([rel([X, Y], R) | O], [rel([X, Y], R), rel([Y, X], R) | OA]) :-
	make_arcs(O, OA).

% consistent(Doms,CA,TDA,A) is true if
% Doms is a set of domains
% CA is a set of consistent arcs,
% TDA is a list of arcs to do
% A is a list of all arcs
consistent(Doms, CA, TDA, A) :-
	consider(Doms, RedDoms, CA, TDA),
	solutions(RedDoms, A).            

% consider(D0,D1,CA,TDA)
% D0 is the set of initial domains
% D1 is the set of reduced domains
% CA = consistent arcs, 
% TDA = to do arcs
consider(D, D, _, []).
consider(D0, D3, CA, [rel([X, Y], R) | TDA]) :-
%	write([rel([X, Y], R) | TDA]), nl,
	choose(dom(XV, DX), D0, D1), identical(X, XV),
%	write("consider: got X = "), write(X), nl,
	choose(dom(YV, DY), D1, _), identical(Y, YV), !,
%	write("consider: got Y = "), write(Y), nl,
	prune(X, DX, Y, DY, R, NDX),
	consider_imp(NDX, DX, D0, D1, D3, X, Y, R, CA, TDA).

consider_imp(NDX, NDX, D0, D1, D3, X, Y, R, CA, TDA) :-
%	write("arc is consistent (NDX = DX)"), nl,
	!, consider(D0, D3, [rel([X, Y], R) | CA], TDA).
consider_imp(NDX, DX, D0, D1, D3, X, Y, R, CA, TDA) :-
%	write("arc is inconsistent (NDX != DX)"), nl,
	acc_todo(X, Y, CA, CA1, TDA, TDA1),
	consider([dom(X, NDX) | D1], D3, [rel([X, Y], R) | CA1], TDA1).

% prune(X,DX,Y,DY,R,NDX)
% variable X had domain DX
% variable Y has domain DY
% R is a relation on X and Y
% NDX = {X in DX | exists Y such that R(X,Y) is true}
prune(_, [], _, _, _, []).
prune(X, [V | XD], Y, YD, R, XD1) :-
	not(prune_not(X, V, Y, YD, R)), !,
%	write("prune: prune_not failed. Throwing out: "), write(V), nl,
	prune(X, XD, Y, YD, R, XD1). 
prune(X, [V | XD], Y, YD, R, [V | XD1]) :-
%	write("prune: retaining: "), write(V), nl,
	prune(X, XD, Y, YD, R, XD1).

prune_not(X, X, Y, YD, R) :-
	member(Y, YD), call(R).

% acc_todo(X,Y,CA,CA1,TDA,TDA1)
% given variables X and Y,
% updates consistent arcs from CA to CA1 and
% to do arcs from TDA to TDA1
acc_todo(_, _, [], [], TDA, TDA).        
acc_todo(X, Y, [rel([U, V], R) | CA0], [rel([U, V], R) | CA1], TDA0, TDA1) :-
	acc_todo_imp(X, V, Y, U),
%   ( X \== V
%   ; X == V,
%     Y == U),   
	acc_todo(X, Y, CA0, CA1, TDA0, TDA1).
acc_todo(X,Y,[rel([U,V],R)|CA0], CA1,TDA0,[rel([U,V],R)|TDA1]) :-
	identical(X, V),
	not_identical(Y, U),
	acc_todo(X, Y, CA0, CA1, TDA0, TDA1).

acc_todo_imp(X, V, _, _) :-
	not_identical(X, V).
acc_todo_imp(X, V, Y, U) :-
	identical(X, V), identical(Y, U).

% solutions(Doms,Arcs) given a reduced set of
% domains, Dome, and arcs Arcs, solves the CSP.
solutions(Doms, _) :-
	solve_singletons(Doms).
solutions(Doms,A) :-
	select(dom(X, [XV1, XV2 | XVs]), Doms, ODoms),
	split([XV1, XV2 | XVs], DX1, DX2),
	acc_todo(X, _, A, CA, [], TDA),
	solutions_or(X, DX1, DX2, ODoms, CA, TDA, A).

solutions_or(X, DX1, DX2, ODoms, CA, TDA, A) :-
	consistent([dom(X,DX1)|ODoms],CA,TDA,A).
solutions_or(X, DX1, DX2, ODoms, CA, TDA, A) :-
	consistent([dom(X,DX2)|ODoms],CA,TDA,A).

% solve_singletons(Doms) is true if Doms is a
% set of singletom domains, with the variables
% assigned to the unique values in the domain
solve_singletons([]).
solve_singletons([dom(X,[X])|Doms]) :-
   solve_singletons(Doms).

% select(E,L,L1) selects the first element of
% L that matches E, with L1 being the remaining
% elements.
select(D,Doms,ODoms) :-
   remove(D,Doms,ODoms), !.

% choose(E,L,L1) chooses an element of
% L that matches E, with L1 being the remaining
% elements.
choose(D, Doms, ODoms) :-
   remove(D, Doms, ODoms).

% split(L,L1,L2) splits list L into two lists L1 and L2
% with the about same number of elements in each list.
split([], [], []).
split([A], [A], []).
split([A, B | R], [A | R1], [B | R2]) :-
   split(R, R1, R2).

%%%%%%%%%%%%%%%%%%%%%

% stdlib predicates

% It's easier in prolog than native
not(X) :- call(X), !, fail.
not(X).

% remove(E,L,R) is true if E is an element of L and R is the remaining
% elements after E is removed.
remove(E, [E|R], R).
remove(E, [A|L], [A|R]) :-
	remove(E, L, R).

% member/2: an old friend.
member(X, [X|_]).
member(X, [_|Ys]) :- member(X, Ys).

% nth(N,L,E) is true if E is the Nth element of list L
% this requires N to be bound.
nth(1,[E|_],E) :- !.
nth(N,[_|R],E) :-
	is(N1, N - 1),
	nth(N1, R ,E).

%%%%%%%%%%%%%%%%%%%%%

%test_csp(A, B, C) :-
%	write([A, B, C]), nl,
%	csp([dom(A, [1, 2, 3, 4]),
%	     dom(B, [1, 2, 3, 4]),
%	     dom(C, [1, 2, 3, 4])],
%	       [rel([A, B], lt(A, B)), rel([B, C], lt(B, C))]).

crossword(A1,A2,A3,D1,D2,D3) :-     
    cross(A1,A2,A3,D1,D2,D3,
          [[a,d,d], [a,d,o], [a,g,e], [a,g,o], [a,i,d],
           [a,i,l], [a,i,m], [a,i,r], [a,n,d], [a,n,y],
           [a,p,e], [a,p,t], [a,r,c], [a,r,e], [a,r,k],
           [a,r,m], [a,r,t], [a,s,h], [a,s,k], [a,u,k],
           [a,w,e], [a,w,l], [a,y,e], [b,a,d], [b,a,g],
           [b,a,n], [b,a,t], [b,e,e], [b,o,a], [e,a,r],
           [e,e,l], [e,f,t], [f,a,r], [f,a,t], [f,i,t],
           [l,e,e], [o,a,f], [r,a,t], [t,a,r], [t,i,e]
          ]).
           

cross(A1,A2,A3,D1,D2,D3,Words) :-
    csp([dom(A1,Words),
         dom(A2,Words),
         dom(A3,Words),
         dom(D1,Words),
         dom(D2,Words),
         dom(D3,Words)],
        [rel([A1,D1],compatible(A1,1,D1,1)),
         rel([A1,D2],compatible(A1,2,D2,1)),
         rel([A1,D3],compatible(A1,3,D3,1)),
         rel([A2,D1],compatible(A2,1,D1,2)),
         rel([A2,D2],compatible(A2,2,D2,2)),
         rel([A2,D3],compatible(A2,3,D3,2)),
         rel([A3,D1],compatible(A3,1,D1,3)),
         rel([A3,D2],compatible(A3,2,D2,3)),
         rel([A3,D3],compatible(A3,3,D3,3))]).

compatible(Word1,I1,Word2,I2) :- 
	common_letter(Word1,I1,Word2,I2).

common_letter(Word1,I1,Word2,I2) :-
	letter(Word1,I1,Letter),
	letter(Word2,I2,Letter).

letter(Word,I,Letter) :- nth(I,Word,Letter).

% | ?- crossword(A,B,C,D,E,F).

main :-
%	test_csp(D, E, F).
	crossword(A,B,C,D,E,F).
